@@ -134,7 +134,7 @@ struct _GDataClientLoginAuthorizerPrivate {
 #endif
 
 	gchar *username;
-	gchar *password;
+	GDataSecureString password; /* must be allocated by _gdata_service_secure_strdup() */
 
 	/* Mapping from GDataAuthorizationDomain to string? auth_token; auth_token is NULL for domains which aren't authorised at the moment */
 	GHashTable *auth_tokens;
@@ -213,6 +213,10 @@ gdata_client_login_authorizer_class_init (GDataClientLoginAuthorizerClass *klass
 	 * then be set to the password passed to gdata_client_login_authorizer_authenticate(), and a #GObject::notify signal will be emitted. If
 	 * authentication fails, it will be set to %NULL.
 	 *
+	 * If libgdata is compiled with libgnome-keyring support, the password will be stored in non-pageable memory. However, if it is retrieved
+	 * using g_object_get() (or related functions) it will be copied to non-pageable memory and could end up being written to disk. Accessing
+	 * the password using gdata_client_login_authorizer_get_password() will not perform any copies, and so maintains privacy.
+	 *
 	 * Since: 0.9.0
 	 */
 	g_object_class_install_property (gobject_class, PROP_PASSWORD,
@@ -289,7 +293,7 @@ gdata_client_login_authorizer_init (GDataClientLoginAuthorizer *self)
 #else
 	g_static_rec_mutex_init (&(self->priv->mutex));
 #endif
-	self->priv->auth_tokens = g_hash_table_new_full (g_direct_hash, g_direct_equal, g_object_unref, g_free);
+	self->priv->auth_tokens = g_hash_table_new_full (g_direct_hash, g_direct_equal, g_object_unref, (GDestroyNotify) _gdata_service_secure_strfree);
 
 	/* Set up the session */
 	self->priv->session = _gdata_service_build_session ();
@@ -319,7 +323,7 @@ finalize (GObject *object)
 	GDataClientLoginAuthorizerPrivate *priv = GDATA_CLIENT_LOGIN_AUTHORIZER (object)->priv;
 
 	g_free (priv->username);
-	g_free (priv->password);
+	_gdata_service_secure_strfree (priv->password);
 	g_free (priv->client_id);
 	g_hash_table_destroy (priv->auth_tokens);
 #if GLIB_CHECK_VERSION (2, 31, 0)
@@ -357,6 +361,7 @@ get_property (GObject *object, guint property_id, GValue *value, GParamSpec *psp
 #endif
 			break;
 		case PROP_PASSWORD:
+			/* NOTE: This takes a pageable copy of non-pageable memory and thus could result in the password hitting disk. */
 #if GLIB_CHECK_VERSION (2, 31, 0)
 			g_rec_mutex_lock (&(priv->mutex));
 			g_value_set_string (value, priv->password);
@@ -405,7 +410,7 @@ set_property (GObject *object, guint property_id, const GValue *value, GParamSpe
 static void
 process_request (GDataAuthorizer *self, GDataAuthorizationDomain *domain, SoupMessage *message)
 {
-	const gchar *auth_token;
+	GDataConstSecureString auth_token; /* privacy sensitive */
 	GDataClientLoginAuthorizerPrivate *priv = GDATA_CLIENT_LOGIN_AUTHORIZER (self)->priv;
 
 	/* If the domain's NULL, return immediately */
@@ -420,7 +425,7 @@ process_request (GDataAuthorizer *self, GDataAuthorizationDomain *domain, SoupMe
 	g_static_rec_mutex_lock (&(priv->mutex));
 #endif
 
-	auth_token = (const gchar*) g_hash_table_lookup (priv->auth_tokens, domain);
+	auth_token = (GDataConstSecureString) g_hash_table_lookup (priv->auth_tokens, domain);
 
 	if (auth_token != NULL) {
 		/* Ensure that we're using HTTPS: if not, we shouldn't set the Authorization header or we could be revealing the auth token to
@@ -428,8 +433,12 @@ process_request (GDataAuthorizer *self, GDataAuthorizationDomain *domain, SoupMe
 		if (soup_message_get_uri (message)->scheme != SOUP_URI_SCHEME_HTTPS) {
 			g_warning ("Not authorizing a non-HTTPS message with the user's ClientLogin auth token as the connection isn't secure.");
 		} else {
+			/* Ideally, authorisation_header would be allocated in non-pageable memory. However, it's copied by
+			 * soup_message_headers_replace() immediately anyway, so there's not much point. However, we do ensure we zero it out before
+			 * freeing. */
 			gchar *authorisation_header = g_strdup_printf ("GoogleLogin auth=%s", auth_token);
 			soup_message_headers_replace (message->request_headers, "Authorization", authorisation_header);
+			memset (authorisation_header, 0, strlen (authorisation_header));
 			g_free (authorisation_header);
 		}
 	}
@@ -546,9 +555,11 @@ notify_authentication_details_cb (GDataClientLoginAuthorizer *self)
 }
 
 static void
-set_authentication_details (GDataClientLoginAuthorizer *self, const gchar *username, const gchar *password, gboolean is_async)
+set_authentication_details (GDataClientLoginAuthorizer *self, const gchar *username, const gchar *password, GHashTable *new_auth_tokens,
+                            gboolean is_async)
 {
 	GDataClientLoginAuthorizerPrivate *priv = self->priv;
+	GHashTableIter iter;
 
 #if GLIB_CHECK_VERSION (2, 31, 0)
 	g_rec_mutex_lock (&(priv->mutex));
@@ -564,8 +575,23 @@ set_authentication_details (GDataClientLoginAuthorizer *self, const gchar *usern
 		priv->username = g_strdup (username);
 	}
 
-	g_free (priv->password);
-	priv->password = g_strdup (password);
+	_gdata_service_secure_strfree (priv->password);
+	priv->password = _gdata_service_secure_strdup (password);
+
+	/* Transfer all successful auth. tokens to the object-wide auth. token store. */
+	if (new_auth_tokens == NULL) {
+		/* Reset ->auth_tokens to contain no auth. tokens, just the domains. */
+		g_hash_table_iter_init (&iter, priv->auth_tokens);
+
+		while (g_hash_table_iter_next (&iter, NULL, NULL) == TRUE) {
+			g_hash_table_iter_replace (&iter, NULL);
+		}
+	} else {
+		/* Replace the existing ->auth_tokens with the new one, which contains all the shiny new auth. tokens. */
+		g_hash_table_ref (new_auth_tokens);
+		g_hash_table_unref (priv->auth_tokens);
+		priv->auth_tokens = new_auth_tokens;
+	}
 
 #if GLIB_CHECK_VERSION (2, 31, 0)
 	g_rec_mutex_unlock (&(priv->mutex));
@@ -585,12 +611,12 @@ set_authentication_details (GDataClientLoginAuthorizer *self, const gchar *usern
 	}
 }
 
-static gboolean
+static GDataSecureString
 parse_authentication_response (GDataClientLoginAuthorizer *self, GDataAuthorizationDomain *domain, guint status,
                                const gchar *response_body, gint length, GError **error)
 {
-	GDataClientLoginAuthorizerPrivate *priv = self->priv;
-	gchar *auth_start, *auth_end, *auth_token;
+	gchar *auth_start, *auth_end;
+	GDataSecureString auth_token; /* NOTE: auth_token must be allocated using _gdata_service_secure_strdup() and friends */
 
 	/* Parse the response */
 	auth_start = strstr (response_body, "Auth=");
@@ -604,28 +630,18 @@ parse_authentication_response (GDataClientLoginAuthorizer *self, GDataAuthorizat
 		goto protocol_error;
 	}
 
-	auth_token = g_strndup (auth_start, auth_end - auth_start);
+	auth_token = _gdata_service_secure_strndup (auth_start, auth_end - auth_start);
 	if (auth_token == NULL || strlen (auth_token) == 0) {
-		g_free (auth_token);
+		_gdata_service_secure_strfree (auth_token);
 		goto protocol_error;
 	}
 
-#if GLIB_CHECK_VERSION (2, 31, 0)
-	g_rec_mutex_lock (&(priv->mutex));
-	g_hash_table_insert (priv->auth_tokens, g_object_ref (domain), auth_token);
-	g_rec_mutex_unlock (&(priv->mutex));
-#else
-	g_static_rec_mutex_lock (&(priv->mutex));
-	g_hash_table_insert (priv->auth_tokens, g_object_ref (domain), auth_token);
-	g_static_rec_mutex_unlock (&(priv->mutex));
-#endif
-
-	return TRUE;
+	return auth_token;
 
 protocol_error:
 	g_set_error_literal (error, GDATA_SERVICE_ERROR, GDATA_SERVICE_ERROR_PROTOCOL_ERROR,
 	                     _("The server returned a malformed response."));
-	return FALSE;
+	return NULL;
 }
 
 static void
@@ -687,7 +703,7 @@ parse_error_response (GDataClientLoginAuthorizer *self, guint status, const gcha
 	             _("Error code %u when authenticating: %s"), status, response_body);
 }
 
-static gboolean
+static GDataSecureString
 authenticate (GDataClientLoginAuthorizer *self, GDataAuthorizationDomain *domain, const gchar *username, const gchar *password,
               gchar *captcha_token, gchar *captcha_answer, GCancellable *cancellable, GError **error)
 {
@@ -696,9 +712,11 @@ authenticate (GDataClientLoginAuthorizer *self, GDataAuthorizationDomain *domain
 	gchar *request_body;
 	const gchar *service_name;
 	guint status;
-	gboolean retval;
+	GDataSecureString auth_token;
 
-	/* Prepare the request */
+	/* Prepare the request.
+	 * NOTE: At this point, our non-pageable password is copied into a pageable HTTP request structure. We can't do much about this
+	 * except note that the request is transient and so the chance of it getting paged out is low (but still positive). */
 	service_name = gdata_authorization_domain_get_service_name (domain);
 	request_body = soup_form_encode ("accountType", "HOSTED_OR_GOOGLE",
 	                                 "Email", username,
@@ -724,7 +742,7 @@ authenticate (GDataClientLoginAuthorizer *self, GDataAuthorizationDomain *domain
 	if (status == SOUP_STATUS_CANCELLED) {
 		/* Cancelled (the error has already been set) */
 		g_object_unref (message);
-		return FALSE;
+		return NULL;
 	} else if (status != SOUP_STATUS_OK) {
 		const gchar *response_body = message->response_body->data;
 		gchar *error_start, *error_end, *uri_start, *uri_end, *uri = NULL;
@@ -885,48 +903,39 @@ login_error:
 		g_free (uri);
 		g_object_unref (message);
 
-		return FALSE;
+		return NULL;
 	}
 
 	g_assert (message->response_body->data != NULL);
 
-	retval = parse_authentication_response (self, domain, status, message->response_body->data, message->response_body->length, error);
+	auth_token = parse_authentication_response (self, domain, status, message->response_body->data, message->response_body->length, error);
+
+	/* Zero out the response body to lower the chance of it (with all the juicy passwords and auth. tokens it contains) hitting disk or getting
+	 * leaked in free memory. */
+	memset ((void*) message->response_body->data, 0, message->response_body->length);
 
 	g_object_unref (message);
 
-	return retval;
+	return auth_token;
 
 protocol_error:
 	parse_error_response (self, status, message->reason_phrase, message->response_body->data, message->response_body->length, error);
 
 	g_object_unref (message);
 
-	return FALSE;
+	return NULL;
 }
 
-typedef struct {
-	gchar *username;
-	gchar *password;
-} AuthenticateAsyncData;
-
-static void
-authenticate_async_data_free (AuthenticateAsyncData *self)
-{
-	g_free (self->username);
-	g_free (self->password);
-
-	g_slice_free (AuthenticateAsyncData, self);
-}
-
-static void
-authenticate_thread (GSimpleAsyncResult *result, GDataClientLoginAuthorizer *authorizer, GCancellable *cancellable)
+static gboolean
+authenticate_loop (GDataClientLoginAuthorizer *authorizer, gboolean is_async, const gchar *username, const gchar *password, GCancellable *cancellable,
+                   GError **error)
 {
 	GDataClientLoginAuthorizerPrivate *priv = authorizer->priv;
-	GError *error = NULL;
-	gboolean success = TRUE;
+	gboolean cumulative_success = TRUE;
+	GHashTable *new_auth_tokens;
 	GHashTableIter iter;
 	GDataAuthorizationDomain *domain;
-	AuthenticateAsyncData *data = g_simple_async_result_get_op_res_gpointer (result);
+	GDataSecureString auth_token;
 
 #if GLIB_CHECK_VERSION (2, 31, 0)
 	g_rec_mutex_lock (&(priv->mutex));
@@ -935,19 +944,25 @@ authenticate_thread (GSimpleAsyncResult *result, GDataClientLoginAuthorizer *aut
 #endif
 
 	/* Authenticate and authorize against each of the services registered with the authorizer */
+	new_auth_tokens = g_hash_table_new_full (g_direct_hash, g_direct_equal, g_object_unref, (GDestroyNotify) _gdata_service_secure_strfree);
 	g_hash_table_iter_init (&iter, priv->auth_tokens);
 
 	while (g_hash_table_iter_next (&iter, (gpointer*) &domain, NULL) == TRUE) {
 		GError *authenticate_error = NULL;
 
-		success = authenticate (authorizer, domain, data->username, data->password, NULL, NULL, cancellable, &authenticate_error) && success;
+		auth_token = authenticate (authorizer, domain, username, password, NULL, NULL, cancellable, &authenticate_error);
 
-		/* Only propagate the first error which occurs. */
-		if (success == FALSE && error == NULL) {
-			g_propagate_error (&error, authenticate_error);
+		if (auth_token == NULL && cumulative_success == TRUE) {
+			/* Only propagate the first error which occurs. */
+			g_propagate_error (error, authenticate_error);
 			authenticate_error = NULL;
 		}
 
+		cumulative_success = (auth_token != NULL) && cumulative_success;
+
+		/* Store the auth. token (or lack thereof if authentication failed). */
+		g_hash_table_insert (new_auth_tokens, g_object_ref (domain), auth_token);
+
 		g_clear_error (&authenticate_error);
 	}
 
@@ -958,12 +973,40 @@ authenticate_thread (GSimpleAsyncResult *result, GDataClientLoginAuthorizer *aut
 #endif
 
 	/* Set or clear the authentication details and return now that we're done */
-	if (success == TRUE) {
-		set_authentication_details (authorizer, data->username, data->password, TRUE);
+	if (cumulative_success == TRUE) {
+		set_authentication_details (authorizer, username, password, new_auth_tokens, is_async);
 	} else {
-		set_authentication_details (authorizer, NULL, NULL, TRUE);
+		set_authentication_details (authorizer, NULL, NULL, NULL, is_async);
 	}
 
+	g_hash_table_unref (new_auth_tokens);
+
+	return cumulative_success;
+}
+
+typedef struct {
+	gchar *username;
+	GDataSecureString password; /* NOTE: This must be allocated in non-pageable memory using _gdata_service_secure_strdup(). */
+} AuthenticateAsyncData;
+
+static void
+authenticate_async_data_free (AuthenticateAsyncData *self)
+{
+	g_free (self->username);
+	_gdata_service_secure_strfree (self->password);
+
+	g_slice_free (AuthenticateAsyncData, self);
+}
+
+static void
+authenticate_thread (GSimpleAsyncResult *result, GDataClientLoginAuthorizer *authorizer, GCancellable *cancellable)
+{
+	GError *error = NULL;
+	gboolean success;
+	AuthenticateAsyncData *data = g_simple_async_result_get_op_res_gpointer (result);
+
+	success = authenticate_loop (authorizer, TRUE, data->username, data->password, cancellable, &error);
+
 	g_simple_async_result_set_op_res_gboolean (result, success);
 
 	if (success == FALSE) {
@@ -1005,7 +1048,7 @@ gdata_client_login_authorizer_authenticate_async (GDataClientLoginAuthorizer *se
 
 	data = g_slice_new (AuthenticateAsyncData);
 	data->username = g_strdup (username);
-	data->password = g_strdup (password);
+	data->password = _gdata_service_secure_strdup (password);
 
 	result = g_simple_async_result_new (G_OBJECT (self), callback, user_data, gdata_client_login_authorizer_authenticate_async);
 	g_simple_async_result_set_handle_cancellation (result, FALSE); /* we handle our own cancellation so we can set ::username and ::password */
@@ -1058,6 +1101,9 @@ gdata_client_login_authorizer_authenticate_finish (GDataClientLoginAuthorizer *s
  * If @cancellable is not %NULL, then the operation can be cancelled by triggering the @cancellable object from another thread.
  * If the operation was cancelled, the error %G_IO_ERROR_CANCELLED will be returned.
  *
+ * If the operation errors or is cancelled part-way through, gdata_authorizer_is_authorized_for_domain() is guaranteed to return %FALSE
+ * for all #GDataAuthorizationDomain<!-- -->s, even if authentication has succeeded for some of them already.
+ *
  * A %GDATA_CLIENT_LOGIN_AUTHORIZER_ERROR_BAD_AUTHENTICATION will be returned if authentication failed due to an incorrect username or password.
  * Other #GDataClientLoginAuthorizerError errors can be returned for other conditions.
  *
@@ -1076,46 +1122,13 @@ gboolean
 gdata_client_login_authorizer_authenticate (GDataClientLoginAuthorizer *self, const gchar *username, const gchar *password,
                                             GCancellable *cancellable, GError **error)
 {
-	GDataClientLoginAuthorizerPrivate *priv;
-	gboolean retval = TRUE;
-	GHashTableIter iter;
-	GDataAuthorizationDomain *domain;
-
 	g_return_val_if_fail (GDATA_IS_CLIENT_LOGIN_AUTHORIZER (self), FALSE);
 	g_return_val_if_fail (username != NULL, FALSE);
 	g_return_val_if_fail (password != NULL, FALSE);
 	g_return_val_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable), FALSE);
 	g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
-	priv = self->priv;
-
-#if GLIB_CHECK_VERSION (2, 31, 0)
-	g_rec_mutex_lock (&(priv->mutex));
-#else
-	g_static_rec_mutex_lock (&(priv->mutex));
-#endif
-
-	/* Authenticate and authorize against each of the services registered with the authorizer */
-	g_hash_table_iter_init (&iter, self->priv->auth_tokens);
-
-	while (g_hash_table_iter_next (&iter, (gpointer*) &domain, NULL) == TRUE) {
-		retval = authenticate (self, domain, username, password, NULL, NULL, cancellable, error) && retval;
-	}
-
-#if GLIB_CHECK_VERSION (2, 31, 0)
-	g_rec_mutex_unlock (&(priv->mutex));
-#else
-	g_static_rec_mutex_unlock (&(priv->mutex));
-#endif
-
-	/* Set or clear the authentication details */
-	if (retval == TRUE) {
-		set_authentication_details (self, username, password, FALSE);
-	} else {
-		set_authentication_details (self, NULL, NULL, FALSE);
-	}
-
-	return retval;
+	return authenticate_loop (self, FALSE, username, password, cancellable, error);
 }
 
 static void
@@ -1278,6 +1291,10 @@ gdata_client_login_authorizer_get_username (GDataClientLoginAuthorizer *self)
  *
  * It is not safe to call this while an authentication operation is ongoing.
  *
+ * If libgdata is compiled with libgnome-keyring support, the password will be stored in non-pageable memory. Since this function doesn't return
+ * a copy of the password, the returned value is guaranteed to not hit disk. It's advised that any copies of the password made in client programs
+ * also use non-pageable memory.
+ *
  * Return value: the password of the currently authenticated user, or %NULL
  *
  * Since: 0.9.0
