@@ -46,6 +46,7 @@
 
 #ifdef HAVE_GNOME
 #include <libsoup/soup-gnome-features.h>
+#include <gnome-keyring-memory.h>
 #endif /* HAVE_GNOME */
 
 #include "gdata-service.h"
@@ -2002,7 +2003,98 @@ debug_handler (const char *log_domain, GLogLevelFlags log_level, const char *mes
 static void
 soup_log_printer (SoupLogger *logger, SoupLoggerLogLevel level, char direction, const char *data, gpointer user_data)
 {
-	g_debug ("%c %s", direction, data);
+	gboolean filter_data;
+	gchar *_data = NULL;
+
+	filter_data = (_gdata_service_get_log_level () > GDATA_LOG_NONE && _gdata_service_get_log_level () < GDATA_LOG_FULL_UNREDACTED) ? TRUE : FALSE;
+
+	if (filter_data == TRUE) {
+		/* Filter out lines which look like they might contain usernames, passwords or auth. tokens. */
+		if (direction == '>' && g_str_has_prefix (data, "Authorization: GoogleLogin ") == TRUE) {
+			_data = g_strdup ("Authorization: GoogleLogin <redacted>");
+		} else if (direction == '>' && g_str_has_prefix (data, "Authorization: OAuth ") == TRUE) {
+			_data = g_strdup ("Authorization: OAuth <redacted>");
+		} else if (direction == '<' && g_str_has_prefix (data, "Set-Cookie: ") == TRUE) {
+			_data = g_strdup ("Set-Cookie: <redacted>");
+		} else if (direction == '<' && g_str_has_prefix (data, "Location: ") == TRUE) {
+			/* Looks like:
+			 * "Location: https://www.google.com/calendar/feeds/default/owncalendars/full?gsessionid=sBjmp05m5i67exYA51XjDA". */
+			SoupURI *uri;
+			gchar *_uri;
+			GHashTable *params;
+
+			uri = soup_uri_new (data + strlen ("Location: "));
+
+			if (uri->query != NULL) {
+				params = soup_form_decode (uri->query);
+
+				/* strdup()s are necessary because the hash table's set up to free keys. */
+				if (g_hash_table_lookup (params, "gsessionid") != NULL) {
+					g_hash_table_insert (params, (gpointer) g_strdup ("gsessionid"), (gpointer) "<redacted>");
+				}
+
+				soup_uri_set_query_from_form (uri, params);
+				g_hash_table_destroy (params);
+			}
+
+			_uri = soup_uri_to_string (uri, FALSE);
+			_data = g_strconcat ("Location: ", _uri, NULL);
+			g_free (_uri);
+
+			soup_uri_free (uri);
+		} else if (direction == '<' && g_str_has_prefix (data, "SID=") == TRUE) {
+			_data = g_strdup ("SID=<redacted>");
+		} else if (direction == '<' && g_str_has_prefix (data, "LSID=") == TRUE) {
+			_data = g_strdup ("LSID=<redacted>");
+		} else if (direction == '<' && g_str_has_prefix (data, "Auth=") == TRUE) {
+			_data = g_strdup ("Auth=<redacted>");
+		} else if (direction == '>' && g_str_has_prefix (data, "accountType=") == TRUE) {
+			/* Looks like: "> accountType=HOSTED%5FOR%5FGOOGLE&Email=[e-mail address]&Passwd=[plaintex password]"
+			               "&service=[service name]&source=ytapi%2DGNOME%2Dlibgdata%2D444fubtt%2D0". */
+			GHashTable *params = soup_form_decode (data);
+
+			/* strdup()s are necessary because the hash table's set up to free keys. */
+			if (g_hash_table_lookup (params, "Email") != NULL) {
+				g_hash_table_insert (params, (gpointer) g_strdup ("Email"), (gpointer) "<redacted>");
+			}
+			if (g_hash_table_lookup (params, "Passwd") != NULL) {
+				g_hash_table_insert (params, (gpointer) g_strdup ("Passwd"), (gpointer) "<redacted>");
+			}
+
+			_data = soup_form_encode_hash (params);
+
+			g_hash_table_destroy (params);
+		} else if (direction == '<' && g_str_has_prefix (data, "oauth_token=") == TRUE) {
+			/* Looks like: "< oauth_token=4%2FI-WU7sBzKk5GhGlQUF8a_TCZRnb7&oauth_token_secret=qTTTJg3no25auiiWFerzjW4I"
+			               "&oauth_callback_confirmed=true". */
+			GHashTable *params = soup_form_decode (data);
+
+			/* strdup()s are necessary because the hash table's set up to free keys. */
+			if (g_hash_table_lookup (params, "oauth_token") != NULL) {
+				g_hash_table_insert (params, (gpointer) g_strdup ("oauth_token"), (gpointer) "<redacted>");
+			}
+			if (g_hash_table_lookup (params, "oauth_token_secret") != NULL) {
+				g_hash_table_insert (params, (gpointer) g_strdup ("oauth_token_secret"), (gpointer) "<redacted>");
+			}
+
+			_data = soup_form_encode_hash (params);
+
+			g_hash_table_destroy (params);
+		} else {
+			/* Nothing to redact. */
+			_data = g_strdup (data);
+		}
+	} else {
+		/* Don't dupe the string. */
+		_data = (gchar*) data;
+	}
+
+	/* Log the data. */
+	g_debug ("%c %s", direction, _data);
+
+	if (filter_data == TRUE) {
+		g_free (_data);
+	}
 }
 
 /**
@@ -2023,7 +2115,7 @@ _gdata_service_get_log_level (void)
 		const gchar *envvar = g_getenv ("LIBGDATA_DEBUG");
 		if (envvar != NULL)
 			level = atoi (envvar);
-		level = MIN (MAX (level, 0), GDATA_LOG_FULL);
+		level = MIN (MAX (level, 0), GDATA_LOG_FULL_UNREDACTED);
 	}
 
 	return level;
@@ -2042,7 +2134,7 @@ _gdata_service_get_log_level (void)
 SoupSession *
 _gdata_service_build_session (void)
 {
-	SoupSession *session = soup_session_sync_new_with_options (SOUP_SESSION_SSL_CA_FILE, CA_CERTS, NULL);
+	SoupSession *session = soup_session_sync_new ();
 
 #ifdef HAVE_GNOME
 	soup_session_add_feature_by_type (session, SOUP_TYPE_GNOME_FEATURES_2_26);
@@ -2054,6 +2146,7 @@ _gdata_service_build_session (void)
 		SoupLogger *logger;
 
 		switch (_gdata_service_get_log_level ()) {
+			case GDATA_LOG_FULL_UNREDACTED:
 			case GDATA_LOG_FULL:
 				level = SOUP_LOGGER_LOG_BODY;
 				break;
@@ -2115,3 +2208,83 @@ gdata_service_set_locale (GDataService *self, const gchar *locale)
 	self->priv->locale = g_strdup (locale);
 	g_object_notify (G_OBJECT (self), "locale");
 }
+
+/*
+ * _gdata_service_secure_strdup:
+ * @str: string (which may be in pageable memory) to be duplicated, or %NULL
+ *
+ * Duplicate a string into non-pageable memory (if libgdata has been compiled with HAVE_GNOME) or just fall back to g_strdup() (if libgdata hasn't).
+ * Passing %NULL to this function will cause %NULL to be returned.
+ *
+ * Strings allocated using this function must be freed using _gdata_service_secure_strfree().
+ *
+ * Return value: non-pageable copy of @str, or %NULL
+ * Since: 0.11.0
+ */
+GDataSecureString
+_gdata_service_secure_strdup (const gchar *str)
+{
+#ifdef HAVE_GNOME
+	return gnome_keyring_memory_strdup (str);
+#else /* if !HAVE_GNOME */
+	return g_strdup (str);
+#endif /* !HAVE_GNOME */
+}
+
+/*
+ * _gdata_service_secure_strndup:
+ * @str: string (which may be in pageable memory) to be duplicated, or %NULL
+ * @n_bytes: maximum number of bytes to copy from @str
+ *
+ * Duplicate at most @n_bytes bytes from @str into non-pageable memory. See _gdata_service_secure_strdup() for more information; this function is just
+ * a version of that with the same semantics as strndup().
+ *
+ * Return value: non-pageable copy of at most the first @n_bytes bytes of @str, or %NULL
+ * Since: 0.11.0
+ */
+GDataSecureString
+_gdata_service_secure_strndup (const gchar *str, gsize n_bytes)
+{
+#ifdef HAVE_GNOME
+	gsize str_len;
+	GDataSecureString duped_str;
+
+	if (str == NULL) {
+		return NULL;
+	}
+
+	str_len = MIN (strlen (str), n_bytes);
+	duped_str = (GDataSecureString) gnome_keyring_memory_alloc (str_len + 1);
+	strncpy (duped_str, str, str_len);
+	*(duped_str + str_len) = '\0';
+
+	return duped_str;
+#else /* if !HAVE_GNOME */
+	return g_strndup (str, n_bytes);
+#endif /* !HAVE_GNOME */
+}
+
+/*
+ * _gdata_service_secure_strfree:
+ * @str: a string to free, or %NULL
+ *
+ * Free a string which was allocated securely using _gdata_service_secure_strdup().
+ * Passing %NULL to this function is safe.
+ *
+ * Since: 0.11.0
+ */
+void
+_gdata_service_secure_strfree (GDataSecureString str)
+{
+#ifdef HAVE_GNOME
+	gnome_keyring_memory_free (str);
+#else /* if !HAVE_GNOME */
+	/* Poor man's approximation to non-pageable memory: the best we can do is ensure that we don't leak it in free memory.
+	 * This can't guarantee that it hasn't hit disk at some point, but does mean it can't hit disk in future. */
+	if (str != NULL) {
+		memset (str, 0, strlen (str));
+	}
+
+	g_free (str);
+#endif /* !HAVE_GNOME */
+}
